///###////////////////////////////////////////////////////////////////////////
//
// Burton Computer Corporation
// http://www.burton-computer.com
// $Id: SpamFilter.cc,v 1.36 2004/01/12 15:42:09 bburton Exp $
//
// Copyright (C) 2000 Burton Computer Corporation
// ALL RIGHTS RESERVED
//
// This program is open source software; you can redistribute it
// and/or modify it under the terms of the Q Public License (QPL)
// version 1.0. Use of this software in whole or in part, including
// linking it (modified or unmodified) into other programs is
// subject to the terms of the QPL.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// Q Public License for more details.
//
// You should have received a copy of the Q Public License
// along with this program; see the file LICENSE.txt.  If not, visit
// the Burton Computer Corporation or CoolDevTools web site
// QPL pages at:
//
//    http://www.burton-computer.com/qpl.html
//

#include <memory>
#include <stdexcept>
#include <algorithm>
#include <cmath>
#include "LockFile.h"
#include "SpamFilter.h"

static const double MIN_PROB = 0.00001;
static const double MAX_PROB = 0.99999;
static const double NEARLY_ZERO = 0.0000001;
static const int MAX_COUNT_FOR_SORT = 20;
static const int COUNT_SORT_DIVISOR = 3;
static const double PROB_SORT_MULTIPLE = 100000.0;
static const double EXT_ARRAY_THRESHOLD = 0.4999;
static const string LOCK_FILENAME("lock");

typedef vector<Token *>::size_type vtokindex_t;

SpamFilter::SpamFilter()
  : m_termsForScore(27),
    m_defaultMinWordCount(3),
    m_maxWordRepeats(2),
    m_extendTopTerms(false),
    m_newWordScore(0.3),
    m_minDistanceForScore(0.0),
    m_minArraySize(0),
    m_waterCounts(false)
{
  setScoreMode(SCORE_NORMAL);
}

SpamFilter::~SpamFilter()
{
  close();
}

void SpamFilter::lock(const File &raw_db_file,
                      bool read_only)
{
  File db_file(FrequencyDB::removeTypePrefix(raw_db_file.getPath()));
  File lock_file(db_file.parent(), LOCK_FILENAME);
  m_lock.set(new LockFile(lock_file));
  m_lock->lock(read_only ? LockFD::SHARED_LOCK : LockFD::EXCLUSIVE_LOCK);
}

void SpamFilter::open(const File &db_file,
                      bool read_only)
{
  lock(db_file, read_only);
  if (!m_db.open(db_file.getPath(), read_only)) {
    throw runtime_error("unable to open words database");
  }
}

void SpamFilter::open(const File &shared_db_file,
                      const File &private_db_file,
                      bool read_only)
{
  lock(private_db_file, read_only);
  if (!m_db.open(shared_db_file.getPath(), private_db_file.getPath(), read_only)) {
    throw runtime_error("unable to open words databases");
  }
}

void SpamFilter::close(bool abandon_writes)
{
  if (m_db.isOpen()) {
    if (!abandon_writes) {
      m_db.flush();
    }
    m_db.close();
  }
  m_lock.clear();
}

double SpamFilter::computeRatio(int count,
                                int total_count)
{
  if (count == 0) {
    return NEARLY_ZERO;
  }

  if (count >= total_count) {
    return MAX_PROB;
  }

  return (double)count / (double)total_count;
}

double SpamFilter::scoreTerm(int good_count,
                             int spam_count,
                             int good_message_count,
                             int spam_message_count)
{
  good_count *= m_goodBias;

  if (m_waterCounts) {
    if (good_count == 0 || spam_count == 0) {
      ++good_count;
      ++spam_count;
    }
  }

  int count = good_count + spam_count;

  double score = m_newWordScore;
  if (count >= m_minWordCount) {
    double good_ratio = computeRatio(good_count, good_message_count);
    double spam_ratio = computeRatio(spam_count, spam_message_count);
    double denom = good_ratio + spam_ratio;
    if (denom > 0.0) {
      score = spam_ratio / denom;
    }
  }

  return max(MIN_PROB, min(MAX_PROB, score));
}

double SpamFilter::scoreToken(Token *tok,
                              int good_message_count,
                              int spam_message_count)
{
  assert(tok->getCount() > 0);

  int good_count, spam_count;
  m_db.getWordCounts(tok->getWord(), good_count, spam_count);
  tok->setDBGoodCount(good_count);
  tok->setDBSpamCount(spam_count);
  if (tok->getDBTotalCount() < 0 && is_debug) {
    cerr << "WARNING: token " << tok->getWord() << " has negative count" << endl;
  }

  double score = scoreTerm(tok->getDBGoodCount(), tok->getDBSpamCount(), good_message_count, spam_message_count);
  if (score > 0.5) {
    tok->setCommonRatio(computeRatio(tok->getDBGoodCount(), good_message_count));
  } else {
    tok->setCommonRatio(computeRatio(tok->getDBSpamCount(), spam_message_count));
  }

  return score;
}

inline bool non_zero(double v)
{
  if (v > 0) {
    return v > NEARLY_ZERO;
  } else {
    return v < -NEARLY_ZERO;
  }
}

// Returns "within document frequency" for sort.  Uses a limit on max
// value for sanity and also uses a ratio of the count to keep small
// differences (3 vs. 2) from having a disproportionate impact on the
// sort.
inline int wdf_for_sort(int count)
{
  return min(MAX_COUNT_FOR_SORT, count / COUNT_SORT_DIVISOR);
}

// Returns probability rounded to 5 decimal places to "band" results for sorting.
// This prevents tiny differences from having excessive impact on results.
inline double rounded_prob(double prob)
{
  return floor(prob * PROB_SORT_MULTIPLE);
}

static int token_qsort_criterion(Token **t1p, Token **t2p)
{
  Token *t1 = *t1p;
  Token *t2 = *t2p;

  double dist = rounded_prob(t1->getDistanceFromMean()) - rounded_prob(t2->getDistanceFromMean());
  if (non_zero(dist)) {
    return (dist > 0) ? -1 : 1;
  }

  // Break ties with in message word count if possible.  Use a limit on count to
  // impose some sanity when the counts are ridiculously big.
  int diff = wdf_for_sort(t1->getCount()) - wdf_for_sort(t2->getCount());
  if (diff != 0) {
    return -diff;
  }

  // next use lower (less spammy) score to break tie if possible
  dist = rounded_prob(t1->getScore()) - rounded_prob(t2->getScore());
  if (non_zero(dist)) {
    return (dist > 0) ? 1 : -1;
  }

  // Break ties with overall word count if possible.
  dist = t1->getCommonRatio() - t2->getCommonRatio();
  if (non_zero(dist)) {
    return (dist > 0) ? -1 : 1;
  }

  // Next use absolute (not rounded) distance from mean
  dist = t1->getDistanceFromMean() - t2->getDistanceFromMean();
  if (non_zero(dist)) {
    return (dist > 0) ? -1 : 1;
  }

  // as a last resort sort alphabetically
  return t1->getWord().compare(t2->getWord());
}

void SpamFilter::getSortedTokens(const Message &msg,
                                 int max_tokens,
                                 vector<Token *> &tokens)
{
  int good_message_count, spam_message_count;
  m_db.getMessageCounts(good_message_count, spam_message_count);

  tokens.clear();
  for (int i = 0; i < msg.getTokenCount(); ++i) {
    Token *tok = msg.getToken(i);
    assert(tok);

    double score = scoreToken(tok, good_message_count, spam_message_count);
    if (score > 0.0) {
      tok->setScore(score);
      tokens.push_back(tok);
    } else {
      tok->setScore(0.0);
    }
  }

  if (tokens.size() == 0) {
    return;
  }

  // STL sort is bugged so we have to use qsort!
  vtokindex_t num_tokens = tokens.size();
  NewPtr<Token *> sorted(new Token *[num_tokens]);
  for (vtokindex_t i = 0; i < num_tokens; ++i) {
    sorted.get()[i] = tokens[i];
  }

  qsort(sorted.get(), num_tokens, sizeof(Token *),
        (int (*) (const void *, const void *))token_qsort_criterion);

  if (is_debug) {
    for (vtokindex_t i = 0; i < num_tokens; ++i) {
      Token *tok = sorted.get()[i];
      cerr << "SORTED " << tok->getWord()
           << " count " << tok->getCount()
           << " score " << tok->getScore()
           << " dist " << tok->getDistanceFromMean()
           << " good " << tok->getDBGoodCount()
           << " spam " << tok->getDBSpamCount()
           << " ratio " << tok->getCommonRatio()
           << endl;
    }
  }

  tokens.clear();
  vtokindex_t max_size = (vtokindex_t)max_tokens;
  for (vtokindex_t i = 0; i < max_size && i < num_tokens; ++i) {
    Token *tok = sorted.get()[i];
    if (tok->getDistanceFromMean() >= m_minDistanceForScore || i < m_minArraySize) {
      tokens.push_back(tok);
    }
  }
}

void SpamFilter::computeScoreProducts(Message &msg,
                                      double &inner,
                                      double &outer,
                                      double &num_terms)
{
  vector<Token *> tokens;
  getSortedTokens(msg, msg.getTokenCount(), tokens);

  msg.clearTopTokens();

  outer = 1.0;
  inner = 1.0;
  num_terms = 0;
  for (vector<Token *>::const_iterator i = tokens.begin(); i != tokens.end(); ++i) {
    Token *tok = *i;

    if ((num_terms >= m_termsForScore) &&
        (!m_extendTopTerms || (tok->getDistanceFromMean() < EXT_ARRAY_THRESHOLD))) {
      break;
    }

    int count = tok->getCount();
    double score = tok->getScore();

    msg.addTopToken(tok);

    assert(count > 0);

    int times = min(m_maxWordRepeats, count);
    if (is_debug) {
      cerr << "** TOKEN " << tok->getWord() << ": score " << score << ": times " << times << endl;
    }

    num_terms += times;
    while (times-- > 0) {
      inner = inner * score;
      outer = outer * (1.0 - score);
    }
  }
}

double SpamFilter::normalScoreMessage(Message &msg)
{
  double inner, outer, num_terms;
  computeScoreProducts(msg, inner, outer, num_terms);

  double score = inner / (inner + outer);
  if (is_nan(score)) {
    score = 0.5;
  }

  if (is_debug) {
    cerr << "** INNER " << inner << ": OUTER " << outer << ": SCORE "
         << score << endl;
  }

  return score;
}

//
// Robinson's simple algorithm.  Refer to his article for details:
// http://radio.weblogs.com/0101454/stories/2002/09/16/spamDetection.html
//
double SpamFilter::alt1ScoreMessage(Message &msg)
{
  double inner, outer, num_terms;
  computeScoreProducts(msg, inner, outer, num_terms);

  double p = 1.0 - pow(outer, 1.0 / num_terms);
  double q = 1.0 - pow(inner, 1.0 / num_terms);
  double s = (p - q) / (p + q);
  s = (s + 1.0) / 2.0;

  if (is_debug) {
    cerr << "SCORE P " << p << ": Q " << q << ": SCORE " << s << endl;
  }

  return s;
}

double SpamFilter::scoreMessage(Message &msg)
{
  switch (m_scoreMode) {
  case SCORE_ALT1:
    return alt1ScoreMessage(msg);

  default:
    return normalScoreMessage(msg);
  }
}

double SpamFilter::scoreMessage(Message &msg,
                                ScoreMode_t mode)
{
  if (m_scoreMode == mode) {
    return scoreMessage(msg);
  }

  ScoreMode_t old_mode = m_scoreMode;
  setScoreMode(mode);
  double score = scoreMessage(msg);
  setScoreMode(old_mode);
  return score;
}

void SpamFilter::ensureGoodMessage(const Message &msg,
                                   bool force_update)
{
  bool is_spam = false;
  if (m_db.containsMessage(msg, is_spam) && is_spam) {
    // remove from database if previously classified as spam
    m_db.removeMessage(msg);
  }
  m_db.addMessage(msg, false, force_update);
}

void SpamFilter::ensureSpamMessage(const Message &msg,
                                   bool force_update)
{
  bool is_spam = false;
  if (m_db.containsMessage(msg, is_spam) && !is_spam) {
    // remove from database if previously classified as good
    m_db.removeMessage(msg);
  }
  m_db.addMessage(msg, true, force_update);
}
